import os
from threading import Lock
import hashlib
import shutil
import subprocess
import sched, time
import threading
from collections import deque

class GitJobDir(object):
    def __init__(s, basedir=None, maxdirs=4):
        s.basedir = basedir
        s.maxdirs = maxdirs
        s.lock = Lock()
        s.use_counts = {}
        s.unused = deque()

        try:
            os.mkdir(basedir)
        except FileExistsError:
            pass

    def dirkey(repo, commit, extra=None):
        _str = "%s::%s::%s" % (repo, commit, extra or "")
        return hashlib.md5(_str.encode("utf-8")).hexdigest()

    def path(s, dirkey):
        return os.path.join(s.basedir, dirkey)

    def get(s, repo, commit, extra=None):
        with s.lock:

            _dir = s.path(GitJobDir.dirkey(repo, commit, extra))

            _users = s.use_counts.get(_dir)
            if _users == None:
                if s.maxdirs == 0:
                    if not s.clean_unused():
                        print("gitjobdir: could not get free jobdir slot")
                        return None
                try:
                    s.checkout(repo, commit, extra)
                    _users = 0
                    s.maxdirs -= 1
                except subprocess.CalledProcessError:
                    return None

            _users += 1
            s.use_counts[_dir] = _users

            return _dir

    def clean_unused(s):
        while True:
            try:
                _dir = s.unused.popleft()
            except IndexError:
                break

            if s.use_counts.get(_dir):
                continue

            s.use_counts.pop(_dir, None)

            GitJobDir.clean_dir(_dir)
            s.maxdirs += 1
            return True

    def release(s, _dir):
        with s.lock:
            use_count = s.use_counts.get(_dir)
            if use_count == 0 or use_count == None:
                print("GitJobDir: warning: release() on unused job dir!")
            else:
                use_count -= 1
                if use_count == 0:
                    print("GitJobDir: last user of %s gone." % _dir)
                    s.unused.append(_dir)

                s.use_counts[_dir] = use_count

    def clean_dir(_dir):
        print("gitjobdir: cleaning directory", _dir)
        shutil.rmtree(_dir)

    def checkout(s, repo, commit, extra):
        target_path = s.path(GitJobDir.dirkey(repo, commit, extra))
        subprocess.check_call(["git", "cache", "clone", repo, commit, target_path])

    def cleanup(s):
        for _dir, v in s.use_counts.items():
            GitJobDir.clean_dir(_dir)

def print_sth():
    print("sth")

if __name__=="__main__":
    gjd = GitJobDir("/tmp/gitjobdir", maxdirs=1)

    _dira = gjd.get("http://github.com/RIOT-OS/RIOT", "c879154d144a349f890ab74ca8e0c70ded359de8")
    print("got", _dira)
    _dirb = gjd.get("http://github.com/RIOT-OS/RIOT", "96ef13dc9801595f4aec8763bd9c36278b5789e9")
    print("got", _dirb)
    print("releasing first")
    gjd.release(_dira)
    if not _dirb:
        print("trying second again", _dirb)
        _dirb = gjd.get("http://github.com/RIOT-OS/RIOT", "96ef13dc9801595f4aec8763bd9c36278b5789e9")
        print("got", _dirb)

    print("releasing 2nd")
    gjd.release(_dirb)

    _exclusive = gjd.get("http://github.com/RIOT-OS/RIOT", "96ef13dc9801595f4aec8763bd9c36278b5789e9", "TEST")
    gjd.release(_exclusive)

    gjd.cleanup()



#!/usr/bin/env python3
import json
import sys
import threading
import random
import time
import socket
import traceback

import cmdserver
from dwq import Job, Disque

from gitjobdir import GitJobDir

def worker(n, cmd_server_pool, gitjobdir):
    print("worker %2i: started" % n)

    my_queues = sys.argv[1:] or ["default"]

    try:
        while True:
            jobs = Job.get(my_queues)
            for job in jobs:
                print("worker %2i: got job %s from queue %s" % (n, job.job_id, job.queue_name))

                try:
                    repo = job.body["repo"]
                    commit = job.body["commit"]
                    command = job.body["command"]
                except KeyError:
                    print("worker %2i: invalid job json body" % n)
                    job.done({ "status" : "error", "output" : "worker.py: invalid job description" })
                    continue

                exclusive = None
                try:
                    options = job.body["options"]
                    if options.get("jobdir") or "" == "exclusive":
                        exclusive = str(random.random())
                except KeyError:
                    pass

                workdir = gitjobdir.get(repo, commit, exclusive)
                if not workdir:
                    job.nack()

                handle = cmd_server_pool.runcmd(command, cwd=workdir, shell=True)
                output, result = handle.wait()

                print("worker %2i: result:" % n, result)
                job.done({ "status" : result, "output" : output, "worker" : hostname })
                gitjobdir.release(workdir)

    except Exception as e:
        print("worker %2i: uncaught exception")
        traceback.print_exc()
        time.sleep("1")

workers = 4
hostname = "unknown"

def main():
    cmd_server_pool = cmdserver.CmdServerPool(workers)
    gitjobdir = GitJobDir("/tmp/wd", workers)

    hostname = socket.gethostname()

    Disque.connect(["localhost:7711"])

    for n in range(1, workers + 1):
        threading.Thread(target=worker, args=(n, cmd_server_pool, gitjobdir), daemon=True).start()

    try:
        while True:
            time.sleep(10)
    except KeyboardInterrupt:
        gitjobdir.cleanup()



import json
from pydisque.client import Client

disque = None

class Disque(object):
    def connect(servers):
        global disque
        disque = Client(servers)
        disque.connect()

class Job(object):
    def __init__(s, job_id, body, queue_name):
        s.job_id = job_id
        s.body = body
        s.queue_name = queue_name
        s.status_queues = body.get("status_queues") or []

    def get(queues):
        global disque
        jobs = []
        _jobs = disque.get_job(queues)
        for queue_name, job_id, json_body in _jobs:
            queue_name = queue_name.decode("ascii")
            job_id = job_id.decode("ascii")
            body = json.loads(json_body.decode("utf-8"))
            jobs.append(Job(job_id, body, queue_name))

        return jobs

    def working(s, status):
        global disque
        disque.working(s.job_id)

    def done(s, result):
        for queue in s.status_queues:
            if queue=="$jobid":
                queue = s.job_id
            print("worker: reporting done to %s" % queue)
            disque.add_job(queue, json.dumps({"job_id" : s.job_id, "state" : "done", "result" : result }))

        disque.ack_job(s.job_id)

    def add(queue, body, status_queues=None, **kwargs):
        body["status_queues"] = status_queues or []

        json_body = json.dumps(body)
        return disque.add_job(queue, json_body, **kwargs).decode("ascii")

    def nack(s):
        disque.nack_job(s.job_id)

    def cancel(s):
        disque.del_job(s.job_id)

    def wait(queue):
        _jobs = disque.get_job([queue])
        for queue_name, job_id, json_body in _jobs:
            queue_name = queue_name.decode("ascii")
            job_id = job_id.decode("ascii")
            body = json.loads(json_body.decode("utf-8"))
            disque.fast_ack(job_id)
            return body

    def cancel_all(job_ids):
        print("canceling all jobs...")
        for job_id in job_ids:
            disque.del_job(job_id)
        print("done.")




#!/usr/bin/env python3

import json
import random
import sys
import time
import argparse

from dwq import Job, Disque

def parse_args():
    parser = argparse.ArgumentParser(prog='dwqc', description='dwq: disque-based work queue')

    parser.add_argument('-q', '--queue', type=str,
            help='queue name for jobs (default: \"default\")', default="default")
    parser.add_argument('--version', action='version', version='%(prog)s 0.1')

    parser.add_argument('-r', "--repo", help='git repository to work on', type=str, required=True)
    parser.add_argument('-c', "--commit", help='git commit to work on', type=str, required=True)
    parser.add_argument('-v', "--verbose", help='enable status output', action="store_true" )
    parser.add_argument('-P', "--progress", help='enable progress output', action="store_true" )
    parser.add_argument('-s', "--stdin", help='read from stdin', action="store_true" )
    parser.add_argument('-Q', "--quiet", help='don\'t print command output', action="store_true" )
    parser.add_argument('command', type=str, nargs='?')

    return parser.parse_args()

def create_body(repo, commit, command, options=None):
    body = { "repo" : repo, "commit" : commit, "command" : command }
    if options:
        body["options"] = options

    return body

def queue_job(jobs_set, queue, body, status_queues):
    job_id = Job.add(queue, body, status_queues)
    jobs_set.add(job_id)

def vprint(*args, **kwargs):
    global verbose
    if verbose:
        print(*args, **kwargs)

verbose = False

def main():
    args = parse_args()

    job_queue = args.queue

    Disque.connect(["localhost:7711"])

    status_queue = "status_%s" % random.random()
    verbose = args.verbose

    try:
        jobs = set()
        if args.command and not args.stdin:
            queue_job(jobs, job_queue, create_body(args.repo, args.commit, args.command), [status_queue])
            result = Job.wait(status_queue)
            print(result["result"]["output"], end="")
            sys.exit(result["result"]["status"])
        else:
            for line in sys.stdin:
                line = line.rstrip()
                if args.stdin:
                    command = args.command.replace("${1}", line)
                else:
                    command = line

                tmp = command.split("###")
                command = tmp[0]
                options = {}
                if len(tmp) > 1:
                    options = json.loads(tmp[1])

                job_id = queue_job(jobs, job_queue, create_body(args.repo, args.commit, command, options), [status_queue])
                vprint("client: job %s command=\"%s\" sent." % (job_id, command))

        total = len(jobs)
        done = 0
        failed = 0
        passed = 0
        while jobs:
            job = Job.wait(status_queue)
            try:
                jobs.remove(job["job_id"])
                done += 1
                if args.progress:
                    vprint("\033[F\033[K", end="")
                vprint("client: job %s done. result=%s" % (job["job_id"], job["result"]["status"]), end="\n\n")
                if not args.quiet:
                    print(job["result"]["output"], end="")
                if job["result"]["status"] in { 0, "0", "pass" }:
                    passed += 1
                else:
                    failed += 1
            except KeyError:
                vprint("client: ignoring unknown job result from id=%s" % job["job_id"])

            finally:
                if args.progress:
                    print("")
            if args.progress:
                print("\033[F\033[Kclient: %s of %s jobs done (%s passed, %s failed.) " % (done, total, passed, failed), end="\r")

        if failed > 0:
            sys.exit(1)
    except KeyboardInterrupt:
        Job.cancel_all(jobs)
        sys.exit(1)
